/*
								+----------------------------------+
								|                                  |
								|   ***  Feature def loader  ***   |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|            Load_XML.h            |
								|                                  |
								+----------------------------------+
*/

#ifndef __FEATURE_XML_LOADER_INCLUDED
#define __FEATURE_XML_LOADER_INCLUDED

struct TWeakClassifier {
	int n_alpha_num;
	float *p_alpha; // alphas for different rank differences

	int n_position_x;
	int n_position_y; // offset of feature, in pixels

	int n_block_a;
	int n_block_b; // zero-based indices of blocks whose ranks are to be evaluated

	int n_block_width;
	int n_block_height; // size of convolution blocks, in pixels

	int n_width;
	int n_height; // size of feature, in blocks

	/*
	 *	inline TWeakClassifier::TWeakClassifier()
	 *		- default constructor
	 */
	TWeakClassifier();

	/*
	 *	inline TWeakClassifier::~TWeakClassifier()
	 *		- default destructor
	 */
	~TWeakClassifier();

	/*
	 *	bool TWeakClassifier::operator =(const TWeakClassifier &r_t_classifier)
	 *		- copy operator
	 *		- returns true on success, false on failure
	 */
	bool operator =(const TWeakClassifier &r_t_classifier);
};

struct TAdaBoostClassifier {
	float f_final_thresh; // final decision threshold

	float f_cutoff_thresh; // cutoff threshold (early calculation termination; unused right now)

	int n_weak_classifier_num;
	TWeakClassifier *p_weak_classifier; // weak classifiers

	char *p_s_description; // contents of "description" field

	/*
	 *	TAdaBoostClassifier::TAdaBoostClassifier()
	 *		- default constructor
	 */
	TAdaBoostClassifier();

	/*
	 *	inline TAdaBoostClassifier::TAdaBoostClassifier(const TAdaBoostClassifier &r_t_classifier)
	 *		- copy-constructor
	 */
	inline TAdaBoostClassifier(const TAdaBoostClassifier &r_t_classifier)
		:p_s_description(0), p_weak_classifier(0)
	{
		Copy(r_t_classifier);
	}

	/*
	 *	TAdaBoostClassifier::~TAdaBoostClassifier()
	 *		- default destructor
	 */
	~TAdaBoostClassifier();

	/*
	 *	inline bool TAdaBoostClassifier::operator =(const TAdaBoostClassifier &r_t_classifier)
	 *		- copy operator
	 *		- returns true on success, false on failure
	 */
	inline bool operator =(const TAdaBoostClassifier &r_t_classifier)
	{
		return Copy(r_t_classifier);
	}

	/*
	 *	bool TAdaBoostClassifier::Copy(const TAdaBoostClassifier &r_t_classifier)
	 *		- copies contents of r_t_classifier to this
	 *		- returns true on success, false on failure
	 */
	bool Copy(const TAdaBoostClassifier &r_t_classifier);

	/*
	 *	bool TAdaBoostClassifier::Load_XML(const char *p_s_filename)
	 *		- loads classifier from XML (Michal Hradis's format)
	 *		- returns true on success, false on failure
	 *		- note original contents of this classifier are deleted
	 */
	bool Load_XML(const char *p_s_filename);
};

#endif //__FEATURE_XML_LOADER_INCLUDED
