#include <vector>
#include "../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <algorithm>
#include <stdlib.h>
#include <string.h>
#include "../berLame_src/OpenGL20.h"
#include "../berLame_src/OpenGLDrv.h"
#include "../berLame_src/OpenGLState.h"
#include "../berLame_src/Texture.h"
#include "../berLame_src/RenderBuffer2.h"
#include "../berLame_src/Shader2.h"
#include "../berLame_src/ShaderSys.h"
#include "../berLame_src/Timer.h"
#include "OpenGL20Logo.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

class CAdHocShaderInfo : public TShaderInfo {
	TShaderInfo::TShader m_t_high;

public:
	CAdHocShaderInfo(const char *p_s_shader_name, bool b_fragment_shader,
		const char *p_s_source_code, const char *p_s_sampler_names = 0)
	{
		p_s_name = (char*)p_s_shader_name;
		b_fragment = b_fragment_shader; // vertex or fragment?
		b_shader_target = true; // "shader", not "program"
		p_low = 0; // no program
		p_high = &m_t_high; // my shader

		m_t_high.p_s_source_code = (char*)p_s_source_code;
		m_t_high.n_sampler_num = 0;
		if(p_s_sampler_names) {
			for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
			   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0)
				++ m_t_high.n_sampler_num;
			// count samplers

			m_t_high.p_sampler = new TShaderInfo::TShader::TSamplerInfo[m_t_high.n_sampler_num];
			_ASSERTE(m_t_high.p_sampler);
			// alloc samplers

			m_t_high.n_sampler_num = 0;
			for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
			   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0) {
				char *p_s_str;
				if(strchr(p_s_sampler, '|')) {
					int n_size = strchr(p_s_sampler, '|') - p_s_sampler;
					if(p_s_str = new char[n_size + 1]) {
						strncpy(p_s_str, p_s_sampler, n_size);
						p_s_str[n_size] = 0;
					}
				} else
					p_s_str = p_s_Alloc(p_s_sampler);
				_ASSERTE(p_s_str);
				// get a single sampler name

				m_t_high.p_sampler[m_t_high.n_sampler_num].p_s_name = p_s_str;
				m_t_high.p_sampler[m_t_high.n_sampler_num].n_resource_index = m_t_high.n_sampler_num;
				// assign samplers

				++ m_t_high.n_sampler_num;
			}
			// assign sampler names and indices
		} else
			m_t_high.p_sampler = 0;
	}

	~CAdHocShaderInfo()
	{
		m_t_high.p_s_source_code = 0;
		// don't want to be deleted by TShaderInfo::TShader destructor as it's static data

		p_s_name = 0;
		p_high = 0;
		// don't want them to be deleted by TShaderInfo destructor as they are static data
	}
};

static HWND h_wnd;
static CGLDriver *p_gl_driver = 0;
static CGLState *p_state = 0;

bool Create_Window(const char *p_s_name, int n_width, int n_height)
{
	WNDCLASSEX t_wnd_class;
	t_wnd_class.cbSize = sizeof(WNDCLASSEX);
	t_wnd_class.style = CS_HREDRAW | CS_VREDRAW;
	t_wnd_class.lpfnWndProc = DefWindowProc;
	t_wnd_class.cbClsExtra = 0;
	t_wnd_class.cbWndExtra = 0;
	t_wnd_class.hInstance = GetModuleHandle(NULL);
	t_wnd_class.hIcon = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	t_wnd_class.hCursor = LoadCursor(GetModuleHandle(NULL), IDC_ARROW);
	t_wnd_class.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	t_wnd_class.lpszMenuName = 0;
	t_wnd_class.lpszClassName = p_s_name;
	t_wnd_class.hIconSm = LoadIcon(GetModuleHandle(NULL), IDI_APPLICATION);
	if(!RegisterClassEx(&t_wnd_class))
		return false;
	// create window class

	RECT t_rect;
	t_rect.left = t_rect.top = 0;
	t_rect.right = n_width;
	t_rect.bottom = n_height;
	if(!AdjustWindowRect(&t_rect, WS_BORDER | WS_CAPTION, false))
		return false;
	// calculate window size to contain client area with desired resolution

	if(!(h_wnd = CreateWindow(p_s_name, p_s_name,
	   WS_BORDER | WS_CAPTION, CW_USEDEFAULT, CW_USEDEFAULT,
	   t_rect.right - t_rect.left, t_rect.bottom - t_rect.top,
	   NULL, NULL, GetModuleHandle(NULL), NULL)))
		return false;
	// create window

	ShowWindow(h_wnd, SW_SHOW);
	UpdateWindow(h_wnd);
	// show & update window

	return true;
}

bool Init_OpenGL() // vraci true, pokud se podarilo. pod linuxem pouziva glut, pod windows pouzije COpenGLDriver z berLame
{
	if(p_state)
		return true;

	if(!Create_Window("OpenGL test", 150, 57))
		return false;
	// create window

	if(!(p_gl_driver = new CGLDriver()))
		return false;
	if(!p_gl_driver->Init(h_wnd, 150, 57, 24, 0)) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// init open-gl

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext()))) {
		delete p_gl_driver;
		p_gl_driver = 0;
		return false;
	}
	// create state guard

	CGLRenderBuffer_FBO::b_Supported(); // gets function pointers as well
	CGLExtensionHandler::n_GetGL12FuncPointers();
	CGLExtensionHandler::n_GetGL13FuncPointers();
	CGLExtensionHandler::n_GetMultitextureFuncPointers(); // opengl1.3 as well
	CGLExtensionHandler::n_GetGL14FuncPointers();
	CGLExtensionHandler::n_GetGL15FuncPointers();
	CGLExtensionHandler::n_GetGL20FuncPointers();
	// get extension function pointers

	unsigned __int32 *p_gl_logo;
	if((p_gl_logo = COpenGL20Logo::p_DecompressImage())) {
		/*glDrawPixels(n_opengl_logo_2_0_sm_jpg_width, n_opengl_logo_2_0_sm_jpg_height,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);*/ // upside-down
		CGLTexture_2D opengl_tex(p_state, COpenGL20Logo::n_Width(),
			COpenGL20Logo::n_Height(), GL_RGBA8, false, 0, GL_TEXTURE_2D,
			GL_BGRA, GL_UNSIGNED_BYTE, p_gl_logo);
		delete[] p_gl_logo;

		p_state->DisableDepthTest();
		p_state->DisableCullFace();
		p_state->Color3f(1, 1, 1);

		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();

		opengl_tex.Bind_Enable(p_state);
		glBegin(GL_QUADS);
		glTexCoord2f(0, 1);
		glVertex2f(-1, -1);
		glTexCoord2f(0, 0);
		glVertex2f(-1,  1);
		glTexCoord2f(1, 0);
		glVertex2f( 1,  1);
		glTexCoord2f(1, 1);
		glVertex2f( 1, -1);
		glEnd();

		p_gl_driver->Blit();
	}
	// load OpenGL logo

	return true;
}

bool Shutdown_OpenGL() // ukonci OpenGL
{
	if(!p_state)
		return true;

	delete p_state;
	p_state = 0;

	p_gl_driver->Shutdown();
	delete p_gl_driver;
	// shutdown open-gl

	DestroyWindow(h_wnd);
	// destroy window

	return true;
}

struct TTestResults {
	double f_fillrate;
	double f_frame_time;

	TTestResults()
		:f_fillrate(0), f_frame_time(0)
	{}

	void operator +=(const TTestResults &r)
	{
		f_fillrate += r.f_fillrate;
		f_frame_time += r.f_frame_time;
	}

	void operator *=(double f)
	{
		f_fillrate *= f; //f_fillrate;
		f_frame_time *= f; //f_frame_time;
	}

	void Print()
	{
		printf("f_fillrate = %f pix/sec\n", f_fillrate);
		printf("f_frame_time = %f sec\n", f_frame_time);
	}
};

CTimer timer;

bool FillrateTest(int n_tex_size, int n_feature_num, int n_pass_num, TTestResults &r_t_results)
{
	printf("create textures ...\r");

	CGLTexture_2D *p_texture; // source texture with packed pixels
	if(!(p_texture = new CGLTexture_2D(p_state, n_tex_size / 4, n_tex_size, GL_RGBA8, false/*,
	   0, GL_TEXTURE_RECTANGLE_ARB*/))) {
		fprintf(stderr, "error: error creating %d x %d rect texture\n", n_tex_size / 4, n_tex_size);
		return false;
	}
	p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	p_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST);
	p_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
	// setup texture ...

	CGLTexture_2D *p_texture2; // dest texture
	if(!(p_texture2 = new CGLTexture_2D(p_state, n_tex_size, n_tex_size, GL_RGB8, false/*,
	   0, GL_TEXTURE_RECTANGLE_ARB*/))) {
		fprintf(stderr, "error: error creating the other %d x %d texture\n", n_tex_size, n_tex_size);
		delete p_texture;
		return false;
	}
	p_texture->r_Parameters(p_state).Set_Texture_Wrap_S(GL_CLAMP_TO_EDGE);
	p_texture->r_Parameters(p_state).Set_Texture_Wrap_T(GL_CLAMP_TO_EDGE);
	p_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST);
	p_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
	// setup the other texture to render to ...

	/*p_state->BindTexture(GL_TEXTURE_RECTANGLE_ARB, 0);
	p_state->DisableTexture(GL_TEXTURE_RECTANGLE_ARB);*/
	p_state->BindTexture2D(0);
	p_state->DisableTexture2D();
	// unbind & disable textures

	printf("create FBO ...     \r");

	p_state->DisableCullFace();
	p_state->DisableDepthTest();
	p_state->Color3f(1, 1, 1);
	// setup state

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_tex_size, n_tex_size, 1, true, 0, // texture format doesn't matter here
	   false, false, 0, false, false, 0)) || !p_render_buffer->b_Status()) {
		fprintf(stderr, "error: error creating %d x %d FBO\n", n_tex_size, n_tex_size);
		if(p_render_buffer)
			delete p_render_buffer;
		delete p_texture;
		delete p_texture2;
		return false;
	}
	// create FBO

	printf("create shader ...  \r");

	/*CAdHocShaderInfo t_fragment_shader("adaBoost_shader", true,
		"#extension GL_ARB_texture_rectangle : require\n" // not supported on 6600 (texture rectangles are supported, but not in programmable pipeline)
		"uniform samplerRect src;\n"
		"uniform int n_feature_num;\n"
		"void main()\n"
		"{\n"
		"    vec3 v_output;\n"
		"    for(int i = 0; i < n_feature_num; ++ i) {\n"
		"        ivec2 v_texcoord = ivec2(gl_TexCoord[0].xy);\n"
		"        vec2 v_tc = gl_TexCoord[0].xy;\n"
		"        v_tc.x *= .25;\n"
		"        int n_path = v_texcoord.x % 4;\n"
		"        if(n_path == 0) {\n"
		"            v_output += textureRect(src, v_tc + vec2(0, -1)).xyz;\n"
		"            v_output += textureRect(src, v_tc).xyz;\n"
		"            v_output += textureRect(src, v_tc + vec2(0, 1)).xyz;\n"
		"        } else if(n_path == 1) {\n"
		"            v_output += textureRect(src, v_tc + vec2(0, -1)).yzw;\n"
		"            v_output += textureRect(src, v_tc).yzw;\n"
		"            v_output += textureRect(src, v_tc + vec2(0, 1)).yzw;\n"
		"        } else if(n_path == 2) {\n"
		"            v_output += vec3(textureRect(src, v_tc + vec2(0, -1)).zw, textureRect(src, v_tc + vec2(1, -1)).x);\n"
		"            v_output += vec3(textureRect(src, v_tc).zw, textureRect(src, v_tc + vec2(1, 0)).x);\n"
		"            v_output += vec3(textureRect(src, v_tc + vec2(0, 1)).zw, textureRect(src, v_tc + vec2(1, 1)).x);\n"
		"        } else /*if(n_path == 3)* / {\n"
		"            v_output += vec3(textureRect(src, v_tc + vec2(0, -1)).w, textureRect(src, v_tc + vec2(1, -1)).xy);\n"
		"            v_output += vec3(textureRect(src, v_tc).w, textureRect(src, v_tc + vec2(1, 0)).xy);\n"
		"            v_output += vec3(textureRect(src, v_tc + vec2(0, 1)).w, textureRect(src, v_tc + vec2(1, 1)).xy);\n"
		"        }\n" // read 3x3 feature from RGB texture
		"    }\n"
		"    gl_FragColor = vec4(v_output, 1);\n"
		"}\n");*/

	CAdHocShaderInfo t_fragment_shader("adaBoost_shader", true,
		"uniform sampler2D src;\n"
		"uniform int n_feature_num;\n"
		"uniform float f_width;\n"
		"uniform float f_dx, f_dy;\n"
		"\n"
		"float f_LRD(vec2 v_texcoord)\n"
		"{\n"
		"    vec3 v_block0, v_block1, v_block2;\n"
		"    vec2 v_tc = v_texcoord;\n"
		"    v_tc.x *= .25;\n"
		"    int n_path = int(v_texcoord.x * f_width) % 4;\n"
		"    if(n_path < 1) {\n"
		"        if(n_path == 1) {\n"
		"            v_block0 = texture2D(src, v_tc + vec2(0, -f_dy)).yzw;\n"
		"            v_block1 = texture2D(src, v_tc).yzw;\n"
		"            v_block2 = texture2D(src, v_tc + vec2(0, f_dy)).yzw;\n"
		"        } else {\n"
		"            v_block0 = texture2D(src, v_tc + vec2(0, -f_dy)).xyz;\n"
		"            v_block1 = texture2D(src, v_tc).xyz;\n"
		"            v_block2 = texture2D(src, v_tc + vec2(0, f_dy)).xyz;\n"
		"        }\n"
		"    } else /*if(n_path > 1)*/ {\n"
		"        if(n_path == 2) {\n"
		"            v_block0 = vec3(texture2D(src, v_tc + vec2(0, -f_dy)).zw, 0/*texture2D(src, v_tc + vec2(f_dx, -f_dy)).x*/);\n"
		"            v_block1 = vec3(texture2D(src, v_tc).zw, 0/*texture2D(src, v_tc + vec2(f_dx, 0)).x*/);\n"
		"            v_block2 = vec3(texture2D(src, v_tc + vec2(0, f_dy)).zw, 0/*texture2D(src, v_tc + vec2(f_dx, f_dy)).x*/);\n"
		"        } else {\n"
		"            v_block0 = vec3(texture2D(src, v_tc + vec2(0, -f_dy)).w, 0, 0/*texture2D(src, v_tc + vec2(f_dx, -f_dy)).xy*/);\n"
		"            v_block1 = vec3(texture2D(src, v_tc).w, 0, 0/*texture2D(src, v_tc + vec2(f_dx, 0)).xy*/);\n"
		"            v_block2 = vec3(texture2D(src, v_tc + vec2(0, f_dy)).w, 0, 0/*texture2D(src, v_tc + vec2(f_dx, f_dy)).xy*/);\n"
		"        }\n"
		"    }\n" // read 3x3 feature from RGB texture
		"\n"
		"    vec4 v_a = vec4(v_block0, v_block1.x);\n"
		"    vec4 v_b = vec4(v_block2, v_block1.z);\n" // divide eight neighbourhood pixels to two vectors
		"    vec4 v_compare = vec4(lessThan(v_a, v_block1.yyyy));\n"
		"    vec4 v_compare2 = vec4(lessThan(v_b, v_block1.yyyy));\n" // compare two vectors with pixel
		"    return dot(v_compare, v_compare) + dot(v_compare2, v_compare2);\n" // sum up
		"}\n"
		"\n"
		"void main()\n"
		"{\n"
		"    float f_output = .0;\n"
		"    vec3 v_block0, v_block1, v_block2;\n"
		"    vec4 v_a = vec4(v_block0, v_block1.x);\n"
		"    vec4 v_b = vec4(v_block2, v_block1.z);\n" // divide eight neighbourhood pixels to two vectors
		"    vec4 v_compare = vec4(lessThan(v_a, v_block1.yyyy));\n"
		"    vec4 v_compare2 = vec4(lessThan(v_b, v_block1.yyyy));\n" // compare two vectors with pixel
		"    f_output = dot(v_compare, v_compare) + dot(v_compare2, v_compare2);\n" // sum up"
		/*
		"    vec2 v_texcoord = gl_TexCoord[0].xy;\n"
		"    float f_3dx = 3.0 * f_dx, f_3dy = 3.0 * f_dy;\n"
		"    for(int i = 0; i < n_feature_num; ++ i) {\n"
		"        f_output += f_LRD(v_texcoord) - f_LRD(v_texcoord + vec2(f_3dx, 0));\n"
		"        v_texcoord.y += f_3dy;\n"
		"    }\n"*/
		"    gl_FragColor = vec4(f_output, f_output, f_output, 1);\n"
		"}\n");

	CGLShaderObject *p_shader;
	char *p_s_compile_log = 0, *p_s_link_log = 0;
	if(!(p_shader = new CGLShaderObject(0, &t_fragment_shader)) ||
	   !p_shader->Compile_Link(p_state, p_s_compile_log, p_s_link_log)) {
		fprintf(stderr, "error: error creating shader\ncompile log: %slink log: %s",
			(p_s_compile_log)? p_s_compile_log : "(null)\n",
			(p_s_link_log)? p_s_link_log : "(null)\n");
		if(p_s_compile_log)
			delete[] p_s_compile_log;
		if(p_s_link_log)
			delete[] p_s_link_log;
		if(p_shader)
			delete p_shader;
		delete p_render_buffer;
		delete p_texture;
		delete p_texture2;
		return false;
	}
	// create shader

	if(p_s_link_log || p_s_compile_log) {
		fprintf(stderr, "warning: shader compile warning(s)\ncompile log: %slink log: %s",
			(p_s_compile_log)? p_s_compile_log : "(null)\n",
			(p_s_link_log)? p_s_link_log : "(null)\n");
		if(p_s_compile_log)
			delete[] p_s_compile_log;
		if(p_s_link_log)
			delete[] p_s_link_log;
	}

	printf("rendering ...      \r");

	double f_start_time = timer.f_Time();
	bool b_result = false;
	for(int i = 0; i < n_pass_num; ++ i) {
		b_result = false;

		if(!p_render_buffer->Bind()) {
			fprintf(stderr, "error: error binding FBO\n");
			break;
		}
		glViewport(0, 0, p_render_buffer->n_Width(), p_render_buffer->n_Height());
		if(!p_render_buffer->BindColorTexture_2D(*p_texture2, GL_TEXTURE_2D/*RECTANGLE_ARB*/)) {
			fprintf(stderr, "error: error binding texture to FBO\n");
			break;
		}
		// bind first texture for rendering ...

		p_texture->Bind_Enable(p_state);
		// bind second texture as source ...

		p_shader->t_Find_IntParameter("n_feature_num").SetValue(n_feature_num);
		p_shader->t_Find_Parameter("f_width").SetValue(n_tex_size);
		p_shader->t_Find_Parameter("f_dx").SetValue(1.0f / n_tex_size);
		p_shader->t_Find_Parameter("f_dy").SetValue(1.0f / n_tex_size);
		p_shader->Bind(p_state);
		// bind shader

		glBegin(GL_QUADS);
		glTexCoord2f(0, 0);
		glVertex2f(-1, -1);
		glTexCoord2f(1/*n_tex_size*/, 0);
		glVertex2f( 1, -1);
		glTexCoord2f(1/*n_tex_size*/, 1/*n_tex_size*/);
		glVertex2f( 1,  1);
		glTexCoord2f(0, 1/*n_tex_size*/);
		glVertex2f(-1,  1);
		glEnd();
		// draw quad (different texcoords for arb texture rectangles!)

		if(glGetError() != GL_NO_ERROR) {
			fprintf(stderr, "error: glGetError() != GL_NO_ERROR\n");
			break;
		}

		p_shader->Release(p_state);

		if(!p_render_buffer->BindColorTexture_2D(0, GL_TEXTURE_2D/*RECTANGLE_ARB*/)) { // release texture
			fprintf(stderr, "error: error releasing texture from FBO\n");
			break;
		}
		if(!p_render_buffer->Release()) {
			fprintf(stderr, "error: error releasing FBO\n");
			break;
		}

		b_result = true;
	}

	glFlush();
	glFinish();

	if(b_result) {
		double f_frame_time = (timer.f_Time() - f_start_time) / n_pass_num;
		r_t_results.f_frame_time = f_frame_time;
		r_t_results.f_fillrate = (n_tex_size * n_tex_size) / f_frame_time;
	}
	// generate results

	printf("cleanup ...        \r");

	delete p_shader;
	delete p_render_buffer;
	delete p_texture;
	delete p_texture2;

	return b_result;
}

int main()
{
	printf("this demo creates set of textures of various sizes and evaluates\n"
		   "number of LRD features. it does not show any output as it's speed\n"
		   "test only and has no detection ability\nenter number of features (up to 256):\n");
	int n_feature_num;
	scanf("%d", &n_feature_num);
	if(n_feature_num < 0 || n_feature_num > 255) {
		printf("error: invalid number of features (%d)\n", n_feature_num);
		return -1;
	}

	if(!Init_OpenGL()) {
		fprintf(stderr, "error: opengl init failure\n");
		return -1;
	}

	const int n_min_resolution = 64;
	const int n_max_resolution = 2048;
	const int n_test_runs = 100;

	TTestResults t_avg_results;
	int n_test_num = 0;
	for(int n_tex_size = n_min_resolution; n_tex_size <= n_max_resolution; n_tex_size <<= 1) {
		printf("\ntest(n_tex_size = %d, n_pass_num = %d)\n", n_tex_size, n_test_runs);

		TTestResults t_results;
		if(!FillrateTest(n_tex_size, n_feature_num, n_test_runs, t_results)) {
			fprintf(stderr, "error: test failure\n");
			Shutdown_OpenGL();
			return -1;
		}
		// run tests many times, average the results

		t_avg_results += t_results;
		++ n_test_num;

		switch(glGetError()) {
		case GL_NO_ERROR:
			break;
		case GL_INVALID_OPERATION:
			fprintf(stderr, "error: GL_INVALID_OPERATION\n");
			break;
		case GL_INVALID_VALUE:
			fprintf(stderr, "error: GL_INVALID_VALUE\n");
			break;
		/*case GL_NO_MEMORY:
			fprintf(stderr, "error: GL_NO_MEMORY\n");
			break;*/
		default:
			fprintf(stderr, "error: GL_ERROR\n");
			break;
		}
		// GL errors ?

		t_results.Print();
	}

	t_avg_results *= 1.0f / n_test_num;
	// average results

	printf("\nAverage results:\n");
	t_avg_results.Print();

	Shutdown_OpenGL();

	return 0;
}
