/*
								+----------------------------------+
								|                                  |
								|  ***  berLame GLUT window  ***  |
								|                                  |
								|   Copyright  -tHE SWINe- 2007   |
								|                                  |
								|             Main.cpp             |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "../../berLame_src/CallStack.h"
#ifdef WIN32
#include <windows.h> // gl.h needs some typedefs
#endif
#include <GL/gl.h>
#include <GL/glut.h>
#include <GL/glext.h>
#include <math.h>
#include "../../berLame_src/OpenGL20.h"
#include "../../berLame_src/OpenGLState.h"
#include "../../berLame_src/VertexBufferObject.h"
#include "../../berLame_src/Shader2.h"
#include "../../berLame_src/ShaderSys.h"
#include "../../berLame_src/Texture.h"
#include "../../berLame_src/RenderBuffer2.h"
#include "../../berLame_src/TextureUtil.h"
#include "../../berLame_src/Vector.h"

static Vector2f v_mouse(0, 0);
// mouse position in OpenGL coords

static int n_width = 512, n_height = 512;
// window size

static CGLState *p_state = 0;
// OpenGL state guard

void onResize(int n_new_width, int n_new_height)
{
	glViewport(0, 0, n_width = n_new_width, n_height = n_new_height);
	// set new viewport

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// set ortho projection

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// set camera
}

// here we render frame
void onDisplay()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	float f_pixel_x = 2.0f / n_width;
	float f_pixel_y = 2.0f / n_height;
	p_state->Color3f(1, 1, 1);
	glBegin(GL_LINES);
	glVertex2f(v_mouse.x - 2 * f_pixel_x, v_mouse.y);
	glVertex2f(v_mouse.x + 3 * f_pixel_x, v_mouse.y);
	glVertex2f(v_mouse.x, v_mouse.y - 2 * f_pixel_y);
	glVertex2f(v_mouse.x, v_mouse.y + 3 * f_pixel_y);
	glEnd();
	// draw cross under mouse cursor

	glutSwapBuffers();
	// display results
}

static int n_button;

void onMouseClick(int _n_button, int n_state, int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;

	n_button = _n_button;

	switch(_n_button) {
    case GLUT_LEFT_BUTTON:
		break;
    case GLUT_RIGHT_BUTTON:
		break;
    case GLUT_MIDDLE_BUTTON:
		break;
	};
}

void onMouseMotion(int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;

	switch(n_button) {
    case GLUT_LEFT_BUTTON:
		break;
    case GLUT_RIGHT_BUTTON:
		break;
    case GLUT_MIDDLE_BUTTON:
		break;
	};
}

void onPassiveMouseMotion(int n_x, int n_y)
{
	v_mouse.x = (2.0f * n_x) / n_width - 1;
	v_mouse.y = 1 - (2.0f * n_y) / n_height;
	glutPostRedisplay();
}

void onKeyPressed(unsigned char n_key_code, int n_x, int n_y)
{
	if(n_key_code == 27) {
		if(p_state)
			delete p_state;
		exit(0);
	}
}

class CAdHocShaderInfo : public TShaderInfo {
	TShaderInfo::TShader m_t_high;

public:
	CAdHocShaderInfo(const char *p_s_shader_name, bool b_fragment_shader,
		const char *p_s_source_code, const char *p_s_sampler_names = 0);
	~CAdHocShaderInfo();
};

CAdHocShaderInfo::CAdHocShaderInfo(const char *p_s_shader_name, bool b_fragment_shader,
	const char *p_s_source_code, const char *p_s_sampler_names)
{
	p_s_name = (char*)p_s_shader_name;
	b_fragment = b_fragment_shader; // vertex or fragment?
	b_shader_target = true; // "shader", not "program"
	p_low = 0; // no program
	p_high = &m_t_high; // my shader

	m_t_high.p_s_source_code = (char*)p_s_source_code;
	m_t_high.n_sampler_num = 0;
	if(p_s_sampler_names) {
		for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
		   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0)
			++ m_t_high.n_sampler_num;
		// count samplers

		m_t_high.p_sampler = new TShaderInfo::TShader::TSamplerInfo[m_t_high.n_sampler_num];
		_ASSERTE(m_t_high.p_sampler);
		// alloc samplers

		m_t_high.n_sampler_num = 0;
		for(const char *p_s_sampler = p_s_sampler_names; p_s_sampler;
		   p_s_sampler = (strchr(p_s_sampler, '|'))? strchr(p_s_sampler, '|') + 1 : 0) {
			char *p_s_str;
			if(strchr(p_s_sampler, '|')) {
				int n_size = strchr(p_s_sampler, '|') - p_s_sampler;
				if(p_s_str = new char[n_size + 1]) {
					strncpy(p_s_str, p_s_sampler, n_size);
					p_s_str[n_size] = 0;
				}
			} else
				p_s_str = p_s_Alloc(p_s_sampler);
			_ASSERTE(p_s_str);
			// get a single sampler name

			m_t_high.p_sampler[m_t_high.n_sampler_num].p_s_name = p_s_str;
			m_t_high.p_sampler[m_t_high.n_sampler_num].n_resource_index = m_t_high.n_sampler_num;
			// assign samplers

			++ m_t_high.n_sampler_num;
		}
		// assign sampler names and indices
	} else
		m_t_high.p_sampler = 0;
}

CAdHocShaderInfo::~CAdHocShaderInfo()
{
	m_t_high.p_s_source_code = 0;
	// don't want to be deleted by TShaderInfo::TShader destructor as it's static data

	p_s_name = 0;
	p_high = 0;
	// don't want them to be deleted by TShaderInfo destructor as they are static data
}

struct TTexCoordRect {
	float f_min_s, f_min_t;
	float f_max_s, f_max_t;

	TTexCoordRect()
	{}

	TTexCoordRect(float f_min, float f_max)
		:f_min_s(f_min), f_min_t(f_min),
		f_max_s(f_max), f_max_t(f_max)
	{}

	TTexCoordRect(float _f_min_s, float _f_min_t, float _f_max_s, float _f_max_t)
		:f_min_s(_f_min_s), f_min_t(_f_min_t),
		f_max_s(_f_max_s), f_max_t(_f_max_t)
	{}
};

float f_fabs(float f_x) // want to avoid conversion to double and back again
{
	union {
		float f;
		unsigned char c[8]; // required by ANSI
	} t_fi = {f_x};
	t_fi.c[3] &= 0x7f;
	return t_fi.f;
}

bool CalibrateRasterizer(CGLState *p_state, int n_width, int n_height,
	TTexCoordRect &r_texcoords, float f_epsilon, int n_max_steps = 100)
{
	if(glGetError() != GL_NO_ERROR) {
		fprintf(stderr, "error: OpenGL errors occured before ::CalibrateRasterizer()\n");
		return false;
	}

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_width, n_height, 1, false,
	   GL_RGBA16F_ARB, false, false, 0, false, false, 0))) {
		fprintf(stderr, "error: error not enough memory\n");
		return false;
	}
	if(glGetError() != GL_NO_ERROR || !p_render_buffer->b_Status()) {
		fprintf(stderr, "error: error creating render-buffer\n");
		return false;
	}
	// create render buffer

	CAdHocShaderInfo shader_src("display_st", true,
		"void main()\n"
		"{\n"
		"    gl_FragColor = gl_TexCoord[0];\n"
		"}\n");
	CGLShaderObject *p_shader;
	if(!(p_shader = new CGLShaderObject(0, &shader_src))) {
		delete p_render_buffer;
		fprintf(stderr, "error: error not enough memory\n");
		return false;
	}
	char *p_s_compile = 0, *p_s_link = 0;
	bool b_result = p_shader->Compile_Link(p_state, p_s_compile, p_s_link);
	if(p_s_compile) {
		fprintf(stderr, "shader compile log:\n%s\n", p_s_compile);
		delete[] p_s_compile;
	}
	if(p_s_link) {
		fprintf(stderr, "shader link log:\n%s\n", p_s_link);
		delete[] p_s_link;
	}
	// create shader

	glViewport(0, 0, n_width, n_height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// setup new viewport and ortho view

	bool b_return_value = false;
	do {
		if(!b_result || !p_shader->b_Status()) {
			fprintf(stderr, "error: error creating shader\n");
			break;
		}
		// create shader to get raster uv's

		p_render_buffer->Bind();
		p_shader->Bind(p_state);
		// bind render buffer, shader

		p_state->DisableDepthTest();
		p_state->DisableAlphaTest();
		p_state->DisableCullFace();
		p_state->DisableBlend();
		// need all those disabaled

		TTexCoordRect t_desired_coords = r_texcoords;

		for(int i = 0; i < n_max_steps; ++ i) {
			glBegin(GL_QUADS);
			glTexCoord2f(r_texcoords.f_min_s, r_texcoords.f_min_t);
			glVertex2f(-1,  1);
			glTexCoord2f(r_texcoords.f_max_s, r_texcoords.f_min_t);
			glVertex2f( 1,  1);
			glTexCoord2f(r_texcoords.f_max_s, r_texcoords.f_max_t);
			glVertex2f( 1, -1);
			glTexCoord2f(r_texcoords.f_min_s, r_texcoords.f_max_t);
			glVertex2f(-1, -1);
			glEnd();
			// draw fullscreen quad

			float p_coords[8];
			glReadPixels(0, 0, 1, 1, GL_RGBA, GL_FLOAT, p_coords);
			glReadPixels(p_render_buffer->n_Width() - 1, p_render_buffer->n_Height() - 1,
				1, 1, GL_RGBA, GL_FLOAT, p_coords + 4);
			// get corner coordinates

			if(f_fabs(t_desired_coords.f_min_s - p_coords[0]) < f_epsilon &&
			   f_fabs(t_desired_coords.f_max_t - p_coords[1]) < f_epsilon &&
			   f_fabs(t_desired_coords.f_max_s - p_coords[4]) < f_epsilon &&
			   f_fabs(t_desired_coords.f_min_t - p_coords[5]) < f_epsilon) {
				b_return_value = true;
				break;
			}
			// compare texcoords

			const float f_weight = .75f;
			r_texcoords.f_min_s += (t_desired_coords.f_min_s - p_coords[0]) * f_weight;
			r_texcoords.f_max_t += (t_desired_coords.f_max_t - p_coords[1]) * f_weight;
			r_texcoords.f_max_s += (t_desired_coords.f_max_s - p_coords[4]) * f_weight;
			r_texcoords.f_min_t += (t_desired_coords.f_min_t - p_coords[5]) * f_weight;
			// offset texcoords
		}

		if(p_shader->n_Api() == shader_api_Core_shader)
			p_state->BindProgramObject_GL2(0);
		else
			p_state->BindProgramObject(0);
		// release shader

		p_render_buffer->Release();
		// release render buffer
	} while(0);

	delete p_shader;
	delete p_render_buffer;

	if(glGetError() != GL_NO_ERROR) {
		fprintf(stderr, "error: OpenGL errors occured\n");
		return false;
	}

	return b_return_value;
}

bool b_RowEmpty(const float *p_buffer, int n_y, int n_width)
{
	n_width *= 4;
	p_buffer += n_y * n_width + 0; // gl_FragColor.x
	for(const float *p_end = p_buffer + n_width; p_buffer != p_end; p_buffer += 4) {
		if(*p_buffer > .5f)
			return false;
	}
	return true;
}

bool b_ColEmpty(const float *p_buffer, int n_x, int n_y0, int n_y1, int n_width)
{
	if(n_y0 > n_y1)
		return false;
	n_width *= 4;
	p_buffer += n_y0 * n_width + 4 * n_x + 0; // gl_FragColor.x
	for(const float *p_end = p_buffer + n_width * (n_y1 - n_y0); p_buffer != p_end; p_buffer += n_width) {
		if(*p_buffer > .5f)
			return false;
	}
	return true;
}

bool FindTexelCenter(float &r_f_s, float &r_f_t, int n_jitter = 10,
	int n_buffer_size = 256, int n_tex_size = 64)
{
	if(glGetError() != GL_NO_ERROR) {
		fprintf(stderr, "error: OpenGL errors occured before ::FindTexelCenter()\n");
		return false;
	}

	unsigned char *p_texture_data;
	if(!(p_texture_data = new unsigned char[n_tex_size * n_tex_size])) {
		fprintf(stderr, "error: error not enough memory\n");
		return false;
	}
	memset(p_texture_data, 0x00, n_tex_size * n_tex_size * sizeof(unsigned char));
	p_texture_data[n_tex_size / 2 + (n_tex_size / 2) * n_tex_size] = 0xff; // white dot in the center
	// create texture data

	CGLTexture_2D *p_texture;
	p_texture = new CGLTexture_2D(p_state, n_tex_size, n_tex_size,
	   GL_LUMINANCE8, true, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, p_texture_data);
	delete[] p_texture_data;
	if(!p_texture || !p_texture->b_Status()) {
		if(p_texture)
			delete p_texture;
		fprintf(stderr, "error: error creating test texture\n");
		return false;
	}
	p_texture->r_Parameters(p_state).Set_Texture_Mag_Filter(GL_NEAREST);
	p_texture->r_Parameters(p_state).Set_Texture_Min_Filter(GL_NEAREST_MIPMAP_NEAREST);
	// create texture

	CGLRenderBuffer_FBO *p_render_buffer;
	if(!(p_render_buffer = new CGLRenderBuffer_FBO(n_buffer_size, n_buffer_size, 1, false,
	   GL_RGBA16F_ARB, false, false, 0, false, false, 0))) {
		fprintf(stderr, "error: error not enough memory\n");
		return false;
	}
	if(glGetError() != GL_NO_ERROR || !p_render_buffer->b_Status()) {
		fprintf(stderr, "error: error creating render-buffer\n");
		return false;
	}
	// create render buffer

	CAdHocShaderInfo shader_src("display_st", true,
		"uniform sampler2D n_texture;\n"
		"void main()\n"
		"{\n"
		"    gl_FragColor.yzw = gl_TexCoord[0].xyw;\n"
		"    gl_FragColor.x = texture2D(n_texture, gl_TexCoord[0].xy).x;\n"
		"}\n");
	CGLShaderObject *p_shader;
	if(!(p_shader = new CGLShaderObject(0, &shader_src))) {
		delete p_render_buffer;
		fprintf(stderr, "error: error not enough memory\n");
		return false;
	}
	char *p_s_compile = 0, *p_s_link = 0;
	bool b_result = p_shader->Compile_Link(p_state, p_s_compile, p_s_link);
	if(p_s_compile) {
		fprintf(stderr, "p_shader compile log:\n%s\n", p_s_compile);
		delete[] p_s_compile;
	}
	if(p_s_link) {
		fprintf(stderr, "p_shader link log:\n%s\n", p_s_link);
		delete[] p_s_link;
	}
	// create shader

	glViewport(0, 0, n_buffer_size, n_buffer_size);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// setup new viewport and ortho view

	bool b_return_value = false;
	do {
		if(!b_result || !p_shader->b_Status()) {
			fprintf(stderr, "error: error creating shader\n");
			break;
		}
		// create shader to get raster uv's

		float *p_buffer;
		if(!(p_buffer = new float[n_buffer_size * n_buffer_size * 4])) {
			p_render_buffer->Release();
			break;
		}

		p_render_buffer->Bind();
		p_shader->Bind(p_state);
		p_texture->Bind_Enable(p_state);
		// bind render buffer, shader, texture

		p_state->DisableDepthTest();
		p_state->DisableAlphaTest();
		p_state->DisableCullFace();
		p_state->DisableBlend();
		// need all those disabaled

		r_f_s = 0;
		r_f_t = 0;

		bool b_error = false;

		float f_jitter_step = (2.0f / n_buffer_size) / n_jitter;
		float f_jitter_x = -f_jitter_step * .5f * n_jitter;
		for(int n_x = 0; n_x < n_jitter && !b_error; ++ n_x, f_jitter_x += f_jitter_step) {
			float f_jitter_y = 0;
			for(int n_y = 0; n_y < n_jitter; ++ n_y, f_jitter_y += f_jitter_step) {
				glBegin(GL_QUADS);
				glTexCoord2f(0, 0);
				glVertex2f(-1 + f_jitter_x,  1 + f_jitter_y);
				glTexCoord2f(1, 0);
				glVertex2f( 1 + f_jitter_x,  1 + f_jitter_y);
				glTexCoord2f(1, 1);
				glVertex2f( 1 + f_jitter_x, -1 + f_jitter_y);
				glTexCoord2f(0, 1);
				glVertex2f(-1 + f_jitter_x, -1 + f_jitter_y);
				glEnd();
				// draw fullscreen textured quad

				glReadPixels(0, 0, n_buffer_size, n_buffer_size, GL_RGBA, GL_FLOAT, p_buffer);
				// read back buffer

				int n_y0, n_y1;
				for(n_y0 = 0; n_y0 < n_buffer_size; ++ n_y0) {
					if(!b_RowEmpty(p_buffer, n_y0, n_buffer_size))
						break;
				}
				if(n_y0 == n_buffer_size) {
					fprintf(stderr, "error: texel was not found\n");
					b_error = true;
					break;
				}
				for(n_y1 = n_y0; n_y1 < n_buffer_size; ++ n_y1) {
					if(b_RowEmpty(p_buffer, n_y1, n_buffer_size))
						break;
				}
				-- n_y1;
				int n_x0, n_x1;
				for(n_x0 = 0; n_x0 < n_buffer_size; ++ n_x0) {
					if(!b_ColEmpty(p_buffer, n_x0, n_y0, n_y1, n_buffer_size))
						break;
				}
				if(n_x0 == n_buffer_size) {
					fprintf(stderr, "error: texel was not found\n");
					b_error = true;
					break;
				}
				for(n_x1 = n_x0; n_x1 < n_buffer_size; ++ n_x1) {
					if(b_ColEmpty(p_buffer, n_x1, n_y0, n_y1, n_buffer_size))
						break;
				}
				-- n_x1;
				// find white pixels

				const float *p_pixel_min0 = p_buffer + ((n_x0 - 1 + n_buffer_size * (n_y0 - 1)) * 4);
				const float *p_pixel_min1 = p_buffer + ((n_x0 + n_buffer_size * n_y0) * 4);
				const float *p_pixel_max0 = p_buffer + ((n_x1 + n_buffer_size * n_y1) * 4);
				const float *p_pixel_max1 = p_buffer + ((n_x1 + 1 + n_buffer_size * (n_y1 + 1)) * 4);
				// min and max pixel

				float f_min_s = (p_pixel_min0[1] + p_pixel_min1[1]) * .5f;
				float f_min_t = (p_pixel_max0[2] + p_pixel_max1[2]) * .5f;
				float f_max_s = (p_pixel_max0[1] + p_pixel_max1[1]) * .5f;
				float f_max_t = (p_pixel_min0[2] + p_pixel_min1[2]) * .5f;
				// pixel bound texcoords (average white and black
				// neighbour coordinates to get border coordintes)

				r_f_s += (f_max_s - f_min_s) * .5f * n_tex_size;
				r_f_t += (f_max_t - f_min_t) * .5f * n_tex_size;
				// pixel center (1 = pixel size)
			}
		}
		r_f_s = (int)(r_f_s + .5f);
		r_f_t = (int)(r_f_t + .5f); // round-off to obtained precision
		r_f_s /= n_jitter * n_jitter;
		r_f_t /= n_jitter * n_jitter;
		// calculate several jittered tests, average results

		delete[] p_buffer;

		if(p_shader->n_Api() == shader_api_Core_shader)
			p_state->BindProgramObject_GL2(0);
		else
			p_state->BindProgramObject(0);
		// release shader

		p_render_buffer->Release();
		// release render buffer

		b_return_value = !b_error;
	} while(0);

	delete p_shader;
	delete p_render_buffer;

	if(glGetError() != GL_NO_ERROR) {
		fprintf(stderr, "error: OpenGL errors occured\n");
		return false;
	}

	return b_return_value;
}

int main(int n_arg_num, const char **p_arg_list)
{
	glutInit(&n_arg_num, (char**)p_arg_list);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
	glutInitWindowSize(n_width, n_height);
	glutCreateWindow("berLame GLUT window");
	// init OpenGL using GLUT

	if(!(p_state = new CGLState(wglGetCurrentDC(), wglGetCurrentContext()))) {
		fprintf(stderr, "error: not enough memory\n");
		return -1;
	}
	// create OpenGL state guard

	printf("=== CalibrateRasterizer ===\n");

	TTexCoordRect t_texcoords(0, 1);
	if(!CalibrateRasterizer(p_state, 512, 512, t_texcoords, 1e-10 /*.25f / 512*/)) {
		fprintf(stderr, "error: your rasterizer is not precise enough; try\n"
						"       increasing rendering quality in display settings\n");
		return -1;
	}
	// calibrate rasterizer

	printf("\tyour rasterizer needs texcoords %g, %g, %g, %g\n"
		"\tto rasterize 512x512 pixels quad with texcoords 0, 0, 1, 1\n",
		t_texcoords.f_min_s, t_texcoords.f_min_t,
		t_texcoords.f_max_s, t_texcoords.f_max_t);
	// reveal results

	printf("\n=== FindTexelCenter ===\n");

	float f_s, f_t;
	if(!FindTexelCenter(f_s, f_t, 10, 256, 64)) {
		fprintf(stderr, "error: error determining texel center coordinates\n");
		return -1;
	}
	// find texel center

	printf("\tyour samplers have texel centre at %g, %g\n\t(relative to pixel size, accurate to 2 decimal places)\n", f_s, f_t);
	// reveal results

	glutDisplayFunc(onDisplay);
	glutReshapeFunc(onResize);
	glutMouseFunc(onMouseClick);
	glutMotionFunc(onMouseMotion);
	glutPassiveMotionFunc(onPassiveMouseMotion);
	glutKeyboardFunc(onKeyPressed);
	// register our callbacks

	glutMainLoop();
	// begin rendering

	return 0;
}
